import six

from ..utils.is_base_type import is_base_type
from .abstracttype import AbstractTypeMeta
from .options import Options
from .unmountedtype import UnmountedType
from .utils import (get_fields_in_type, yank_fields_from_attrs,
                    get_base_fields, merge)


class InputObjectTypeMeta(AbstractTypeMeta):

    def __new__(cls, name, bases, attrs):
        # Also ensure initialization is only performed for subclasses of
        # InputObjectType
        if not is_base_type(bases, InputObjectTypeMeta):
            return type.__new__(cls, name, bases, attrs)

        options = Options(
            attrs.pop('Meta', None),
            name=name,
            description=attrs.get('__doc__'),
            local_fields=None,
        )

        options.base_fields = get_base_fields(InputObjectType, bases)

        if not options.local_fields:
            options.local_fields = get_fields_in_type(InputObjectType, attrs)
            yank_fields_from_attrs(attrs, options.local_fields)

        options.fields = merge(
            options.base_fields,
            options.local_fields
        )
        return type.__new__(cls, name, bases, dict(attrs, _meta=options))

    def __str__(cls):  # noqa: N802
        return cls._meta.name


class InputObjectType(six.with_metaclass(InputObjectTypeMeta, UnmountedType)):

    @classmethod
    def get_type(cls):
        return cls



from collections import OrderedDict

import six

from ..utils.is_base_type import is_base_type
from .options import Options
from .unmountedtype import UnmountedType

try:
    from enum import Enum as PyEnum
except ImportError:
    from ..utils.enum import Enum as PyEnum


class EnumTypeMeta(type):

    def __new__(cls, name, bases, attrs):
        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        if not is_base_type(bases, EnumTypeMeta):
            return type.__new__(cls, name, bases, attrs)

        options = Options(
            attrs.pop('Meta', None),
            name=name,
            description=attrs.get('__doc__'),
            enum=None,
        )
        if not options.enum:
            options.enum = PyEnum(cls.__name__, attrs)

        new_attrs = OrderedDict(attrs, _meta=options, **options.enum.__members__)
        return type.__new__(cls, name, bases, new_attrs)

    def __prepare__(name, bases, **kwargs):  # noqa: N805
        return OrderedDict()

    def __call__(cls, *args, **kwargs):  # noqa: N805
        if cls is Enum:
            description = kwargs.pop('description', None)
            return cls.from_enum(PyEnum(*args, **kwargs), description=description)
        return super(EnumTypeMeta, cls).__call__(*args, **kwargs)

    def from_enum(cls, enum, description=None):  # noqa: N805
        meta_class = type('Meta', (object,), {'enum': enum, 'description': description})
        return type(meta_class.enum.__name__, (Enum,), {'Meta': meta_class})

    def __str__(cls):  # noqa: N805
        return cls._meta.name


class Enum(six.with_metaclass(EnumTypeMeta, UnmountedType)):

    def get_type(self):
        return type(self)



from .unmountedtype import UnmountedType


class Structure(UnmountedType):
    '''
    A structure is a GraphQL type instance that
    wraps a main type with certain structure.
    '''

    def __init__(self, of_type, *args, **kwargs):
        super(Structure, self).__init__(*args, **kwargs)
        self.of_type = of_type

    def get_type(self):
        return self


class List(Structure):

    def __str__(self):
        return '[{}]'.format(self.of_type)


class NonNull(Structure):

    def __str__(self):
        return '{}!'.format(self.of_type)



from collections import OrderedDict

from .dynamic import Dynamic
from .field import Field
from .inputfield import InputField
from .unmountedtype import UnmountedType


def merge_fields_in_attrs(bases, attrs):
    from ..types import AbstractType, Interface
    inherited_bases = (AbstractType, Interface)
    for base in bases:
        if base in inherited_bases or not issubclass(base, inherited_bases):
            continue
        for name, field in base._meta.fields.items():
            if name in attrs:
                continue
            attrs[name] = field
    return attrs


def merge(*dicts):
    merged = OrderedDict()
    for _dict in dicts:
        merged.update(_dict)
    return merged


def get_base_fields(in_type, bases):
    fields = OrderedDict()
    fields = merge_fields_in_attrs(bases, fields)
    return get_fields_in_type(in_type, fields, order=False)


def unmounted_field_in_type(unmounted_field, type):
    '''
    Mount the UnmountedType dinamically as Field or InputField
    depending on where mounted in.

    ObjectType -> Field
    InputObjectType -> InputField
    '''
    # from ..types.inputobjecttype import InputObjectType
    from ..types.objecttype import ObjectType
    from ..types.interface import Interface
    from ..types.abstracttype import AbstractType
    from ..types.inputobjecttype import InputObjectType

    if issubclass(type, (ObjectType, Interface)):
        return unmounted_field.Field()

    elif issubclass(type, (AbstractType)):
        return unmounted_field
    elif issubclass(type, (InputObjectType)):
        return unmounted_field.InputField()

    raise Exception(
        'Unmounted field "{}" cannot be mounted in {}.'.format(
            unmounted_field, type
        )
    )


def get_field(in_type, value):
    if isinstance(value, (Field, InputField, Dynamic)):
        return value
    elif isinstance(value, UnmountedType):
        return unmounted_field_in_type(value, in_type)


def get_fields_in_type(in_type, attrs, order=True):
    fields_with_names = []
    for attname, value in list(attrs.items()):
        field = get_field(in_type, value)
        if not field:
            continue
        fields_with_names.append((attname, field))

    if not order:
        return OrderedDict(fields_with_names)

    return OrderedDict(sorted(fields_with_names, key=lambda f: f[1]))


def yank_fields_from_attrs(attrs, fields):
    for name in fields.keys():
        del attrs[name]



import inspect

from ..utils.orderedtype import OrderedType


class Dynamic(OrderedType):

    def __init__(self, type, _creation_counter=None):
        super(Dynamic, self).__init__(_creation_counter=_creation_counter)
        assert inspect.isfunction(type)
        self.type = type

    def get_type(self):
        return self.type()
